@@ -1,4 +1,5 @@
-/* Copyright (c) 2003-2007 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,11 +12,13 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <NDBT.hpp>
 #include <NDBT_Table.hpp>
 #include <NDBT_Tables.hpp>
+#include <NdbEnv.h>
 
 /* ******************************************************* */
 //    Define Ndb standard tables 
@@ -23,6 +26,11 @@
 //  USE ONLY UPPERLETTERS IN TAB AND COLUMN NAMES
 /* ******************************************************* */
 
+static const NdbDictionary::Column::StorageType MM=
+    NdbDictionary::Column::StorageTypeMemory;
+static const NdbDictionary::Column::StorageType DD=
+    NdbDictionary::Column::StorageTypeDisk;
+
 /*
  * These are our "official" test tables
  *
@@ -41,33 +49,81 @@ static
 const
 NDBT_Table T1("T1", sizeof(T1Attribs)/sizeof(NDBT_Attribute), T1Attribs);
 
-/* T2 */
 static
 const
 NDBT_Attribute T2Attribs[] = {
-  NDBT_Attribute("KOL1", NdbDictionary::Column::Bigunsigned, 1, true), 
-  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL3", NdbDictionary::Column::Bit, 23),
-  NDBT_Attribute("KOL4", NdbDictionary::Column::Unsigned, 
-		 1, false, true), // Nullable 
-  NDBT_Attribute("KOL5", NdbDictionary::Column::Unsigned)
+  NDBT_Attribute("KOL1", NdbDictionary::Column::Varbinary, 100, true), 
+  NDBT_Attribute("KOL2", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL3", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL4", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL5", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL6", NdbDictionary::Column::Unsigned),
+  NDBT_Attribute("KOL7", NdbDictionary::Column::Unsigned)
 };
+
 static
 const
 NDBT_Table T2("T2", sizeof(T2Attribs)/sizeof(NDBT_Attribute), T2Attribs);
 
-/* T3 */
 static
 const
 NDBT_Attribute T3Attribs[] = {
-  NDBT_Attribute("ID", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("PERSNR", NdbDictionary::Column::Char, 10),
-  NDBT_Attribute("NAME", NdbDictionary::Column::Char, 25),
-  NDBT_Attribute("ADRESS", NdbDictionary::Column::Char, 50),
-  NDBT_Attribute("ADRESS2", NdbDictionary::Column::Char, 
-		 30, false, true), // Nullable
-  NDBT_Attribute("FDELSER", NdbDictionary::Column::Unsigned)
+  NDBT_Attribute("KOL1", NdbDictionary::Column::Varbinary, 100, true), 
+  NDBT_Attribute("KOL00", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL01", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL02", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL03", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL04", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL05", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL06", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL07", NdbDictionary::Column::Varbinary, 25),
+  NDBT_Attribute("KOL08", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL09", NdbDictionary::Column::Varbinary, 25),
+  NDBT_Attribute("KOL10", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL11", NdbDictionary::Column::Varbinary, 25),
+  NDBT_Attribute("KOL12", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL13", NdbDictionary::Column::Varbinary, 25),
+  NDBT_Attribute("KOL14", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL15", NdbDictionary::Column::Longvarbinary, 537),
+  NDBT_Attribute("KOL16", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL17", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL18", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL19", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL20", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL21", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL22", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL23", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL24", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL25", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL26", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL27", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL28", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL29", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL30", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL31", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL32", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL33", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL34", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL35", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL36", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL37", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL38", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL39", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL40", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL41", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL42", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL43", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL44", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL45", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL46", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL47", NdbDictionary::Column::Varbinary, 100),
+  NDBT_Attribute("KOL48", NdbDictionary::Column::Char, 100),
+  NDBT_Attribute("KOL49", NdbDictionary::Column::Varbinary, 100),
+
+  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned),
+  NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned)
 };
+
 static
 const
 NDBT_Table T3("T3", sizeof(T3Attribs)/sizeof(NDBT_Attribute), T3Attribs);
@@ -76,30 +132,18 @@ NDBT_Table T3("T3", sizeof(T3Attribs)/sizeof(NDBT_Attribute), T3Attribs);
 static
 const
 NDBT_Attribute T4Attribs[] = {
-  NDBT_Attribute("REGNR", NdbDictionary::Column::Char, 6, true), 
-  NDBT_Attribute("YEAR", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("OWNER", NdbDictionary::Column::Char, 25),
-  NDBT_Attribute("ADRESS", NdbDictionary::Column::Char, 50),
-  NDBT_Attribute("ADRESS2", NdbDictionary::Column::Char, 
-		 30, false, true), // Nullable
-  NDBT_Attribute("OWNERID", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("CHECKDATE", NdbDictionary::Column::Unsigned)
+  NDBT_Attribute("KOL1", NdbDictionary::Column::Unsigned, 1, true,  false, 0, MM, true), 
+  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned, 1, false, false, 0, MM, true),
+  NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned, 1, false, false, 0, MM, true),
+  NDBT_Attribute("KOL4", NdbDictionary::Column::Unsigned, 1, false, true, 0, MM, true),
+  NDBT_Attribute("KOL5", NdbDictionary::Column::Unsigned, 1, false, false, 0, MM, true)
 };
+
 static
 const
 NDBT_Table T4("T4", sizeof(T4Attribs)/sizeof(NDBT_Attribute), T4Attribs);
 
-/* T5 */
-static
-const
-NDBT_Attribute T5Attribs[] = {
-  NDBT_Attribute("OWNERID", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("REGNR", NdbDictionary::Column::Char, 6, true),
-  NDBT_Attribute("CREATEDDATE", NdbDictionary::Column::Unsigned)
-};
-static
-const
-NDBT_Table T5("T5", sizeof(T5Attribs)/sizeof(NDBT_Attribute), T5Attribs);
+
 
 /* T6 */
 static
@@ -122,140 +166,21 @@ NDBT_Attribute T6Attribs[] = {
   NDBT_Attribute("ATTR11", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("ATTR12", NdbDictionary::Column::Char, 49),
   NDBT_Attribute("ATTR13", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR14", NdbDictionary::Column::Char, 50),
+  NDBT_Attribute("ATTR14", NdbDictionary::Column::Varbinary, 50),
   NDBT_Attribute("ATTR15", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("ATTR16", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("ATTR17", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("ATTR18", NdbDictionary::Column::Char, 257),
   NDBT_Attribute("ATTR19", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("ATTR20", NdbDictionary::Column::Unsigned),
+  NDBT_Attribute("ATTR21", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("ATTR22", NdbDictionary::Column::Varbinary, 255, false, false, 0, MM, true)
 };
+
 static
 const
 NDBT_Table T6("T6", sizeof(T6Attribs)/sizeof(NDBT_Attribute), T6Attribs);
 
-/* T7 */
-static
-const
-NDBT_Attribute T7Attribs[] = {
-  NDBT_Attribute("PK1", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("PK2", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("PK3", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("PK4", NdbDictionary::Column::Varbinary, 123, true), 
-  NDBT_Attribute("ATTR1", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR2", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR3", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR4", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR5", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR6", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR7", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR8", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR9", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR10", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR11", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR12", NdbDictionary::Column::Char, 259),
-  NDBT_Attribute("ATTR13", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR14", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR15", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR16", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR17", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR18", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR19", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("ATTR20", NdbDictionary::Column::Unsigned),
-};
-static
-const
-NDBT_Table T7("T7", sizeof(T7Attribs)/sizeof(NDBT_Attribute), T7Attribs);
-
-/* T8 */
-static
-const
-NDBT_Attribute T8Attribs[] = {
-  NDBT_Attribute("PERSON_ID", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("NAME", NdbDictionary::Column::Varbinary, 255),
-  NDBT_Attribute("ADRESS", NdbDictionary::Column::Longvarbinary, 513),
-  NDBT_Attribute("POSTADRESS", NdbDictionary::Column::Char, 1173),
-  NDBT_Attribute("VALUE", NdbDictionary::Column::Unsigned),
-  
-};
-static
-const
-NDBT_Table T8("T8", sizeof(T8Attribs)/sizeof(NDBT_Attribute), T8Attribs);
-
-/* T9 */
-static
-const
-NDBT_Attribute T9Attribs[] = {
-  NDBT_Attribute("KF_SKAPAD", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("PLATS_ID", NdbDictionary::Column::Char, 2, true),
-  NDBT_Attribute("TNR_SKAPAD", NdbDictionary::Column::Char, 12, true),
-  NDBT_Attribute("DELG_MOT", NdbDictionary::Column::Char, 1, true),
-  NDBT_Attribute("VALUE", NdbDictionary::Column::Unsigned),
-};
-static
-const
-NDBT_Table T9("T9", sizeof(T9Attribs)/sizeof(NDBT_Attribute), T9Attribs);
-
-/* T10 - Long key table */
-static
-const
-NDBT_Attribute T10Attribs[] = {
-  NDBT_Attribute("KOL1", NdbDictionary::Column::Char, 256, true), 
-  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL3", NdbDictionary::Column::Char, 257),
-  NDBT_Attribute("KOL4", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL5", NdbDictionary::Column::Unsigned),
-};
-static
-const
-NDBT_Table T10("T10", sizeof(T10Attribs)/sizeof(NDBT_Attribute), T10Attribs);
-
-
-/* T11 - Primary key is not first attribute */
-static
-const
-NDBT_Attribute T11Attribs[] = {
-  NDBT_Attribute("KOL1", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL4", NdbDictionary::Column::Char, 111),
-  NDBT_Attribute("KOL5", NdbDictionary::Column::Char, 113)
-};
-
-static
-const
-NDBT_Table T11("T11", sizeof(T11Attribs)/sizeof(NDBT_Attribute), T11Attribs);
-
-/* T12 - 16 primary keys */
-static
-const
-NDBT_Attribute T12Attribs[] = {
-  NDBT_Attribute("KOL1", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL4", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL5", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL6", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL7", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL8", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL9", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL10", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL11", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL12", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL13", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL14", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL15", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL16", NdbDictionary::Column::Unsigned, 1, true), 
-  NDBT_Attribute("KOL20", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL30", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL40", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL50", NdbDictionary::Column::Unsigned)
-};
-
-static
-const
-NDBT_Table T12("T12", sizeof(T12Attribs)/sizeof(NDBT_Attribute), T12Attribs);
-
 /* T13 - Long key table */
 static
 const
@@ -271,6 +196,7 @@ static
 const
 NDBT_Table T13("T13", sizeof(T13Attribs)/sizeof(NDBT_Attribute), T13Attribs);
 
+
 /* T14 - 5 primary keys */
 static
 const
@@ -283,7 +209,7 @@ NDBT_Attribute T14Attribs[] = {
   NDBT_Attribute("KOL20", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("KOL30", NdbDictionary::Column::Int),
   NDBT_Attribute("KOL40", NdbDictionary::Column::Float),
-  NDBT_Attribute("KOL50", NdbDictionary::Column::Char, 200)
+  NDBT_Attribute("KOL50", NdbDictionary::Column::Char, 200, false, false, 0, MM, true)
 };
 
 
@@ -291,6 +217,88 @@ static
 const
 NDBT_Table T14("T14", sizeof(T14Attribs)/sizeof(NDBT_Attribute), T14Attribs);
 
+/*
+  T15 - Dynamic attributes.
+  Test many different combinations of attribute types, sizes, and NULLability.
+  Also exersize >32bit dynattr bitmap.
+*/
+static
+const
+NDBT_Attribute T15Attribs[] = {
+  NDBT_Attribute("KOL1", NdbDictionary::Column::Unsigned, 1, true, false, 0, MM, true),
+  NDBT_Attribute("KOL2", NdbDictionary::Column::Varbinary, 100, false, true, 0, MM, true),
+  NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned, 1, false, true, 0, MM, true),
+  NDBT_Attribute("KOL4", NdbDictionary::Column::Int, 1, false, false, 0, MM, true),
+  NDBT_Attribute("KOL5", NdbDictionary::Column::Float, 1, false, true, 0, MM, true),
+  NDBT_Attribute("KOL6", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL7", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL8", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL9", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL10", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL11", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL12", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL13", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL14", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL15", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL16", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL17", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL18", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL19", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL20", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL21", NdbDictionary::Column::Varbinary, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL22", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL23", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL24", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL25", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL26", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL27", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL28", NdbDictionary::Column::Char, 4, false, false),
+  NDBT_Attribute("KOL29", NdbDictionary::Column::Varbinary, 4, false, false),
+  NDBT_Attribute("KOL30", NdbDictionary::Column::Char, 4, false, true, 0, DD),
+  NDBT_Attribute("KOL31", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL32", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("BIT1", NdbDictionary::Column::Bit, 27, false, true, 0, MM, true),
+  NDBT_Attribute("BIT2", NdbDictionary::Column::Bit, 1, false, false, 0, MM, true),
+  NDBT_Attribute("BIT3", NdbDictionary::Column::Bit, 1, false, true, 0, MM, true),
+  NDBT_Attribute("BIT4", NdbDictionary::Column::Bit, 8, false, false, 0, MM, true),
+  NDBT_Attribute("KOL33", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL34", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL35", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL36", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL37", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL38", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL39", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL40", NdbDictionary::Column::Varbinary, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL41", NdbDictionary::Column::Char, 64, false, true, 0, MM, true),
+  NDBT_Attribute("KOL42", NdbDictionary::Column::Char, 4, false, true, 0, MM, true),
+  NDBT_Attribute("KOL43", NdbDictionary::Column::Char, 8, false, true, 0, MM, true),
+  NDBT_Attribute("KOL44", NdbDictionary::Column::Char, 27, false, true, 0, MM, true),
+  NDBT_Attribute("KOL45", NdbDictionary::Column::Char, 64, false, false, 0, MM, true),
+  NDBT_Attribute("KOL46", NdbDictionary::Column::Char, 4, false, false, 0, MM, true),
+  NDBT_Attribute("KOL47", NdbDictionary::Column::Char, 8, false, false, 0, MM, true),
+  NDBT_Attribute("KOL48", NdbDictionary::Column::Char, 27, false, false, 0, MM, true),
+  NDBT_Attribute("KOL49", NdbDictionary::Column::Varbinary, 255, false, false, 0, MM, true),
+  /* This one is for update count, needed by hugoScanUpdate. */
+  NDBT_Attribute("KOL99", NdbDictionary::Column::Unsigned, 1, false, false, 0, MM, true),
+};
+
+static
+const
+NDBT_Table T15("T15", sizeof(T15Attribs)/sizeof(NDBT_Attribute), T15Attribs);
+
+/* Test dynamic bit types when no other varsize/dynamic. */
+static
+const
+NDBT_Attribute T16Attribs[] = {
+  NDBT_Attribute("KOL1", NdbDictionary::Column::Unsigned, 1, true, false),
+  NDBT_Attribute("Kol2", NdbDictionary::Column::Bit, 27, false, true, 0, MM, true),
+  NDBT_Attribute("KOL99", NdbDictionary::Column::Unsigned, 1, false, false),
+};
+
+static
+const
+NDBT_Table T16("T16", sizeof(T16Attribs)/sizeof(NDBT_Attribute), T16Attribs);
+
 /*
   C2 DHCP TABLES, MAYBE THESE SHOULD BE MOVED TO THE UTIL_TABLES?
 */
@@ -391,20 +399,50 @@ NDBT_Attribute D1Attribs[] = {
   NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned),
   NDBT_Attribute("KOL4", NdbDictionary::Column::Char, 233, false, true, 0, NdbDictionary::Column::StorageTypeDisk),
   NDBT_Attribute("KOL5", NdbDictionary::Column::Unsigned),
+  NDBT_Attribute("KOL6", NdbDictionary::Column::Varbinary, 233, false, true, 0, NdbDictionary::Column::StorageTypeDisk),
 };
 static
 const
 NDBT_Table D1("D1", sizeof(D1Attribs)/sizeof(NDBT_Attribute), D1Attribs);
 
+static
+const char* BigVarDefault =
+  "\x80\x1"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2"
+  "KOL7 default in table D2";
+
+static unsigned smallUintDefault = 77;
+
 static
 const
 NDBT_Attribute D2Attribs[] = {
   NDBT_Attribute("KOL1", NdbDictionary::Column::Varbinary, 127, true), 
-  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned, 1, false, false, 0, NdbDictionary::Column::StorageTypeDisk),
+  NDBT_Attribute("KOL2", NdbDictionary::Column::Unsigned, 1, false, false, 0, NdbDictionary::Column::StorageTypeDisk, false,
+                 &smallUintDefault, sizeof(unsigned)),
   NDBT_Attribute("KOL3", NdbDictionary::Column::Unsigned),
-  NDBT_Attribute("KOL4", NdbDictionary::Column::Varbinary, 133),
+  NDBT_Attribute("KOL4", NdbDictionary::Column::Varbinary, 133, false, true, 0, MM, true, 
+                 "\x1E" "A default value for KOL4 in D2", 31),
   NDBT_Attribute("KOL5", NdbDictionary::Column::Char, 199, false, true, 0, NdbDictionary::Column::StorageTypeDisk),
   NDBT_Attribute("KOL6", NdbDictionary::Column::Bit, 21, false, false, 0, NdbDictionary::Column::StorageTypeDisk),
+  NDBT_Attribute("KOL7", NdbDictionary::Column::Longvarbinary, 384, false, true, 0, NdbDictionary::Column::StorageTypeDisk, false, 
+                 BigVarDefault, 386),
+  NDBT_Attribute("KOL8", NdbDictionary::Column::Varbinary, 88, false, true, 0, NdbDictionary::Column::StorageTypeDisk, false, 
+                 "\x1E""A default value for KOL8 in D2", 31)
+
 };
 static
 const
@@ -420,36 +458,17 @@ NDBT_Table *test_tables[]=
   &T2,
   &T3,
   &T4,
-  &T5,
   &T6,
-  &T7,
-  &T8,
-  &T9,
-  &T10,
-  &T11,
-  &T12,
   &T13,
   &T14,
+  &T15,
+  &T16,
   &I1,
   &I2,
   &I3,
   &D1, &D2
 };
 
-struct NDBT_IndexList {
-  const char * m_table;
-  const char ** m_indexes;
-};
-
-static
-const
-NDBT_IndexList indexes[] = {
-  "I1", I1_Indexes, 
-  "I2", I2_Indexes, 
-  "I3", I3_Indexes,
-  0, 0
-};
-
 static
 const
 int numTestTables = sizeof(test_tables)/sizeof(NDBT_Table*);
@@ -756,6 +775,260 @@ const
 int numUtilTables = sizeof(util_tables)/sizeof(NDBT_Table*);
 
 
+/**
+ * Define other test tables that we may create
+ */
+
+/* WIDE_2COL
+ * Single var length key going up to max size of key
+ * Single var length attribute using remaining row space
+ */
+static 
+const
+NDBT_Attribute WIDE_2COL_ATTRIBS[] = {
+  /* Note that we can't have any index on this table as it
+   * has no space for the extra FRAGID the index requires!
+   */
+  NDBT_Attribute("KEY", NdbDictionary::Column::Longvarbinary, 
+                 NDBT_Tables::MaxVarTypeKeyBytes, true), 
+  NDBT_Attribute("ATTR", NdbDictionary::Column::Longvarbinary, 
+                 NDBT_Tables::MaxKeyMaxVarTypeAttrBytes, false)
+};
+
+
+static
+NDBT_Table WIDE_2COL("WIDE_2COL", sizeof(WIDE_2COL_ATTRIBS)/
+                     sizeof(NDBT_Attribute), WIDE_2COL_ATTRIBS);
+
+/* WIDE_2COL_IX
+ * Single var length key going up to max size of key
+ * Single var length attribute using most of remaining row space,
+ * but not all as we need space in the index for FragId
+ */
+static 
+const
+NDBT_Attribute WIDE_2COL_IX_ATTRIBS[] = {
+  NDBT_Attribute("KEY", NdbDictionary::Column::Longvarbinary,
+                 NDBT_Tables::MaxVarTypeKeyBytes, true), 
+  NDBT_Attribute("ATTR", NdbDictionary::Column::Longvarbinary,
+                 NDBT_Tables::MaxKeyMaxVarTypeAttrBytesIndex , false)
+};
+
+static
+NDBT_Table WIDE_2COL_IX("WIDE_2COL_IX", sizeof(WIDE_2COL_IX_ATTRIBS)/
+                        sizeof(NDBT_Attribute), WIDE_2COL_IX_ATTRIBS);
+
+static
+const char* WIDE_2COL_IX_Indexes[] = {
+  "UNIQUE", "ATTR", 0,
+  0};
+
+/* WIDE_MAXKEY_HUGO
+ * Single var length key going up to max size of key
+ * Var length attr going using up remaining space
+ * Two unsigned int columns required by Hugo tools
+ */
+static 
+const
+NDBT_Attribute WIDE_MAXKEY_HUGO_ATTRIBS[] = {
+  /* Note that we can't have any index on this table as it
+   * has no space for the extra FRAGID the index requires!
+   */
+  NDBT_Attribute("KEY", NdbDictionary::Column::Longvarbinary,
+                 NDBT_Tables::MaxVarTypeKeyBytes, true), 
+  NDBT_Attribute("ATTR", NdbDictionary::Column::Longvarbinary,
+                 NDBT_Tables::MaxKeyMaxVarTypeAttrBytes -
+                 NDBT_Tables::HugoOverheadBytes, false),
+  NDBT_Attribute("HUGOID", NdbDictionary::Column::Unsigned,
+                 1, false),
+  NDBT_Attribute("HUGOUPDATE", NdbDictionary::Column::Unsigned,
+                 1, false)
+};
+
+static
+NDBT_Table WIDE_MAXKEY_HUGO("WIDE_MAXKEY_HUGO", sizeof(WIDE_MAXKEY_HUGO_ATTRIBS)/
+                          sizeof(NDBT_Attribute), WIDE_MAXKEY_HUGO_ATTRIBS);
+
+/* WIDE_MAXATTR_HUGO
+ * Single unsigned int key
+ * Var length attr using up remaining space
+ * Two unsigned int columns required by Hugo tools
+ */
+static 
+const
+NDBT_Attribute WIDE_MAXATTR_HUGO_ATTRIBS[] = {
+  NDBT_Attribute("KEY", NdbDictionary::Column::Unsigned,
+                 1, true), 
+  NDBT_Attribute("ATTR", NdbDictionary::Column::Longvarbinary,
+                 NDBT_Tables::MinKeyMaxVarTypeAttrBytes -
+                 NDBT_Tables::HugoOverheadBytes, false),
+  NDBT_Attribute("HUGOID", NdbDictionary::Column::Unsigned,
+                 1, false),
+  NDBT_Attribute("HUGOUPDATE", NdbDictionary::Column::Unsigned,
+                 1, false)
+};
+
+static
+NDBT_Table WIDE_MAXATTR_HUGO("WIDE_MAXATTR_HUGO", sizeof(WIDE_MAXATTR_HUGO_ATTRIBS)/
+                             sizeof(NDBT_Attribute), WIDE_MAXATTR_HUGO_ATTRIBS);
+
+typedef NDBT_Table* (*TableGenerator)(const char* name);
+
+static NDBT_Table* WIDE_MAXKEYMAXCOLS_HUGO= NULL;
+
+static
+NDBT_Table* createMaxKeyMaxColsHugoTabDef(const char* name)
+{
+  if (WIDE_MAXKEYMAXCOLS_HUGO == NULL)
+  {
+    /* Create a wide table with the max num of keys
+     * and the max num of attrs
+     */
+    NdbDictionary::Column* attrs[NDB_MAX_ATTRIBUTES_IN_TABLE];
+    const int buffsize=100;
+    char namebuff[buffsize];
+    Uint32 attrNum=0;
+
+    /* Keys */
+    for (;attrNum < (NDB_MAX_ATTRIBUTES_IN_INDEX - 1); attrNum ++)
+    {
+      BaseString::snprintf(namebuff, buffsize, "K%d", attrNum);
+      attrs[attrNum]= new NDBT_Attribute(namebuff,
+                                         NdbDictionary::Column::Unsigned,
+                                         1, true);
+    }
+    /* Last key uses remaining key space */
+    BaseString::snprintf(namebuff, buffsize, "K%d", attrNum);
+    attrs[attrNum]= new NDBT_Attribute(namebuff,
+                                       NdbDictionary::Column::Char,
+                                       (NDB_MAX_KEYSIZE_IN_WORDS -
+                                        (NDB_MAX_ATTRIBUTES_IN_INDEX -1)) * 4,
+                                       true);
+
+    attrNum ++;
+
+    /* Attributes */
+    for (;attrNum < (NDB_MAX_ATTRIBUTES_IN_TABLE - 1); attrNum ++)
+    {
+      BaseString::snprintf(namebuff, buffsize, "A%d", attrNum);
+      attrs[attrNum]= new NDBT_Attribute(namebuff,
+                                         NdbDictionary::Column::Unsigned,
+                                         1, false);
+    }
+
+    /* Last attr uses remaining attr space */
+    BaseString::snprintf(namebuff, buffsize, "A%d", attrNum);
+    Uint32 sz32 = NDB_MAX_TUPLE_SIZE_IN_WORDS;
+    sz32 -= NDB_MAX_KEYSIZE_IN_WORDS;
+    sz32 -= NDB_MAX_ATTRIBUTES_IN_INDEX;
+    sz32 -= 2 * NDB_MAX_ATTRIBUTES_IN_TABLE;
+    attrs[attrNum]= new NDBT_Attribute(namebuff,
+                                       NdbDictionary::Column::Char, 4 * sz32,
+                                       false);
+
+    WIDE_MAXKEYMAXCOLS_HUGO= new NDBT_Table(name, NDB_MAX_ATTRIBUTES_IN_TABLE,
+                                            attrs);
+
+    /* Free attributes, table will remain until program exit */
+    for (attrNum=0; attrNum < NDB_MAX_ATTRIBUTES_IN_TABLE; attrNum++)
+      delete attrs[attrNum];
+    
+  }
+
+  return WIDE_MAXKEYMAXCOLS_HUGO;
+}
+
+static NDBT_Table* WIDE_MINKEYMAXCOLS_HUGO= NULL;
+
+static
+NDBT_Table* createMinKeyMaxColsHugoTabDef(const char* name)
+{
+  if (WIDE_MINKEYMAXCOLS_HUGO == NULL)
+  {
+    /* Create a wide table with one key and the max number
+     * of attributes
+     */
+    NdbDictionary::Column* attrs[NDB_MAX_ATTRIBUTES_IN_TABLE];
+    const int buffsize=100;
+    char namebuff[buffsize];
+    Uint32 attrNum=0;
+    attrs[attrNum]= new NDBT_Attribute("K1",
+                                       NdbDictionary::Column::Unsigned,
+                                       1, true);
+    attrNum ++;
+
+    /* Attributes */
+    for (;attrNum < (NDB_MAX_ATTRIBUTES_IN_TABLE - 1); attrNum ++)
+    {
+      BaseString::snprintf(namebuff, buffsize, "A%d", attrNum);
+      attrs[attrNum]= new NDBT_Attribute(namebuff,
+                                         NdbDictionary::Column::Unsigned,
+                                         1, false);
+    }
+    
+    /* Last attr uses remaining attr space */
+    BaseString::snprintf(namebuff, buffsize, "A%d", attrNum);
+    attrs[attrNum]= new NDBT_Attribute(namebuff,
+                                       NdbDictionary::Column::Char,
+                                       (NDB_MAX_TUPLE_SIZE_IN_WORDS -
+                                        (NDB_MAX_ATTRIBUTES_IN_TABLE - 1)) * 4,
+                                       false);
+    
+    WIDE_MINKEYMAXCOLS_HUGO= new NDBT_Table(name, NDB_MAX_ATTRIBUTES_IN_TABLE,
+                                            attrs);
+
+    /* Free attributes, table will remain until program exit */
+    for (attrNum=0; attrNum < NDB_MAX_ATTRIBUTES_IN_TABLE; attrNum++)
+      delete attrs[attrNum];
+    
+
+  }
+
+  return WIDE_MINKEYMAXCOLS_HUGO;
+}
+
+// Define array with pointer to other test tables
+
+struct OtherTable
+{
+  const char* name;
+  NDBT_Table* tab;
+  TableGenerator tabGen;
+};
+
+static
+const
+OtherTable other_tables[]=
+{ 
+  {"WIDE_2COL", &WIDE_2COL, NULL},
+  {"WIDE_2COL_IX", &WIDE_2COL_IX, NULL},
+  {"WIDE_MAXKEY_HUGO", &WIDE_MAXKEY_HUGO, NULL},
+  {"WIDE_MAXATTR_HUGO", &WIDE_MAXATTR_HUGO, NULL},
+  {"WIDE_MAXKEYMAXCOLS_HUGO", NULL, createMaxKeyMaxColsHugoTabDef},
+  {"WIDE_MINKEYMAXCOLS_HUGO", NULL, createMinKeyMaxColsHugoTabDef}
+};
+
+static
+const
+int numOtherTables = sizeof(other_tables)/sizeof(OtherTable);
+
+
+/* Secondary indexes for our tables */
+struct NDBT_IndexList {
+  const char * m_table;
+  const char ** m_indexes;
+};
+
+static
+const
+NDBT_IndexList indexes[] = {
+  { "I1", I1_Indexes }, 
+  { "I2", I2_Indexes }, 
+  { "I3", I3_Indexes },
+  { "WIDE_2COL_IX", WIDE_2COL_IX_Indexes },
+  { 0, 0 }
+};
+
 const
 NdbDictionary::Table*
 NDBT_Tables::getTable(const char* _nam){
@@ -777,6 +1050,14 @@ NDBT_Tables::getTable(const char* _nam){
       return util_tables[i];
     }
   }
+  for(i=0; i<numOtherTables; i++){
+    if (strcmp(other_tables[i].name, _nam) == 0){
+      return (other_tables[i].tab != NULL)? 
+        other_tables[i].tab :
+        (*other_tables[i].tabGen)(other_tables[i].name);
+    }
+  }
+
   // TPK_no tables
   // Dynamcially create table vith primary key size
   // set to no
@@ -864,11 +1145,21 @@ NDBT_Tables::create_default_tablespace(Ndb* pNdb)
   NdbDictionary::Dictionary* pDict = pNdb->getDictionary();
 
   int res;
+  Uint32 mb = 8;
+  {
+    char buf[256];
+    if (NdbEnv_GetEnv("UNDOBUFFER", buf, sizeof(buf)))
+    {
+      mb = atoi(buf);
+      ndbout_c("Using %umb dd-undo-buffer", mb);
+    }
+  }
+
   NdbDictionary::LogfileGroup lg = pDict->getLogfileGroup("DEFAULT-LG");
   if (strcmp(lg.getName(), "DEFAULT-LG") != 0)
   {
     lg.setName("DEFAULT-LG");
-    lg.setUndoBufferSize(8*1024*1024);
+    lg.setUndoBufferSize(mb*1024*1024);
     res = pDict->createLogfileGroup(lg);
     if(res != 0){
       g_err << "Failed to create logfilegroup:"
@@ -876,28 +1167,41 @@ NDBT_Tables::create_default_tablespace(Ndb* pNdb)
       return NDBT_FAILED;
     }
   }
+
+  mb = 96;
+  Uint32 files = 13;
+
   {
-    NdbDictionary::Undofile uf = pDict->getUndofile(0, "undofile01.dat");
-    if (strcmp(uf.getPath(), "undofile01.dat") != 0)
+    char buf[256];
+    if (NdbEnv_GetEnv("UNDOSIZE", buf, sizeof(buf)))
     {
-      uf.setPath("undofile01.dat");
-      uf.setSize(32*1024*1024);
-      uf.setLogfileGroup("DEFAULT-LG");
-      
-      res = pDict->createUndofile(uf, true);
-      if(res != 0){
-	g_err << "Failed to create undofile:"
-	      << endl << pDict->getNdbError() << endl;
-	return NDBT_FAILED;
-      }
+      mb = atoi(buf);
+      ndbout_c("Using %umb dd-undo", mb);
     }
   }
+  
   {
-    NdbDictionary::Undofile uf = pDict->getUndofile(0, "undofile02.dat");
-    if (strcmp(uf.getPath(), "undofile02.dat") != 0)
+    char buf[256];
+    if (NdbEnv_GetEnv("UNDOFILES", buf, sizeof(buf)))
     {
-      uf.setPath("undofile02.dat");
-      uf.setSize(32*1024*1024);
+      files = atoi(buf);
+      ndbout_c("Using max %u dd-undo files", files);
+    }
+  }
+  
+  Uint32 sz = 32;
+  while (mb > files * sz)
+    sz += 32;
+
+  for (Uint32 i = 0; i * sz < mb; i++)
+  {
+    char tmp[256];
+    BaseString::snprintf(tmp, sizeof(tmp), "undofile%u.dat", i);
+    NdbDictionary::Undofile uf = pDict->getUndofile(0, tmp);
+    if (strcmp(uf.getPath(), tmp) != 0)
+    {
+      uf.setPath(tmp);
+      uf.setSize(Uint64(sz)*1024*1024);
       uf.setLogfileGroup("DEFAULT-LG");
       
       res = pDict->createUndofile(uf, true);
@@ -908,6 +1212,7 @@ NDBT_Tables::create_default_tablespace(Ndb* pNdb)
       }
     }
   }
+
   NdbDictionary::Tablespace ts = pDict->getTablespace("DEFAULT-TS");
   if (strcmp(ts.getName(), "DEFAULT-TS") != 0)
   {
@@ -922,30 +1227,39 @@ NDBT_Tables::create_default_tablespace(Ndb* pNdb)
       return NDBT_FAILED;
     }
   }
-  
-  {
-    NdbDictionary::Datafile df = pDict->getDatafile(0, "datafile01.dat");
-    if (strcmp(df.getPath(), "datafile01.dat") != 0)
-    {
-      df.setPath("datafile01.dat");
-      df.setSize(64*1024*1024);
-      df.setTablespace("DEFAULT-TS");
-      
-      res = pDict->createDatafile(df, true);
-      if(res != 0){
-	g_err << "Failed to create datafile:"
-	      << endl << pDict->getNdbError() << endl;
-	return NDBT_FAILED;
-      }
-    }
-  }
 
+  mb = 128;
   {
-    NdbDictionary::Datafile df = pDict->getDatafile(0, "datafile02.dat");
-    if (strcmp(df.getPath(), "datafile02.dat") != 0)
+    char buf[256];
+    if (NdbEnv_GetEnv("DATASIZE", buf, sizeof(buf)))
     {
-      df.setPath("datafile02.dat");
-      df.setSize(64*1024*1024);
+      mb = atoi(buf);
+      ndbout_c("Using %umb dd-data", mb);
+    }
+  }
+  
+  sz = 64;
+  files = 13;
+  {
+    char buf[256];
+    if (NdbEnv_GetEnv("DATAFILES", buf, sizeof(buf)))
+    {
+      files = atoi(buf);
+      ndbout_c("Using max %u dd-data files", files);
+    }
+  }
+  
+  while (mb > files * sz)
+    sz += 32;
+  for (Uint32 i = 0; i * sz < mb; i++)
+  {
+    char tmp[256];
+    BaseString::snprintf(tmp, sizeof(tmp), "datafile%u.dat", i);
+    NdbDictionary::Datafile df = pDict->getDatafile(0, tmp);
+    if (strcmp(df.getPath(), tmp) != 0)
+    {
+      df.setPath(tmp);
+      df.setSize(Uint64(sz)*1024*1024);
       df.setTablespace("DEFAULT-TS");
       
       res = pDict->createDatafile(df, true);
@@ -956,7 +1270,7 @@ NDBT_Tables::create_default_tablespace(Ndb* pNdb)
       }
     }
   }
-  
+    
   return NDBT_OK;
 }
 
@@ -999,6 +1313,7 @@ loop:
     if(r == -1){
       if(pNdb->getDictionary()->getNdbError().code == 755)
       {
+	ndbout << "Error: " << pNdb->getDictionary()->getNdbError() << endl;
 	if (create_default_tablespace(pNdb) == 0)
 	{
 	  goto loop;
@@ -1042,7 +1357,7 @@ loop:
 	
 	j++;
 	while(indexes[i].m_indexes[j] != 0){
-	  tmpIndx.addIndexColumn(indexes[i].m_indexes[j]);
+          tmpIndx.addIndexColumn(indexes[i].m_indexes[j]);
 	  j++;
 	}
 	j++;
